### The code below contains the two functions for nonparameteric and parametric nonlinear indicator-level moderation models. The algorithm makes use of functions written in C. The C-code can be found in the supplementary file 'NLILMLVM.c'. The C-code needs
### to be compiled to obtain a .dll file before it can be used. To compile the code the following steps should be taken:
###
### 1. Add R bin folder to PATH
### 2. Install R-tools. 
### 3. Add "c:\Rtools\bin;c:\Rtools\gcc-4.6.3\bin;" to PATH 
### 4. Copy the file mixture.c to the R-bin directory (for example C:\R\R-3.3.2\bin\x64) and ensure that writing permission is enabled for that folder
### 5. In Windows command line go to the R-bin directory (e.g., cd C:\R\R-3.3.2\bin\x64) 
### 6. In Windows command line type in: R CMD SHLIB NLILMLVM.c
### 7. Paste the produced file mixture.dll to your R working directory together with 'NLILMLVM.c'.
###


dyn.load('NLILMLVM.dll') # loading compiled c-code, the file should be in the working directory

library(statmod) # package needed to obtain the Gauss-Hermite quadrature nodes and weights
library(ltm) # package needed to specify the initial values

# The input for the analysis should be:
# Y is the matrix of observed binary indicator
# Z is the matrix of moderators  (The dimensions of Y and Z should be the same)
# h is the bandwidth
# g is how many focal points are used
# Quant is TRUE if equally spaces quantiles of the distribution should be used
# if Quant==FALSE, then a matrix of focal points F needs to be specifies with g rows and as many columns as there are indicators
# nD number of nodes for the Gauss-Hermite quadrature

nonparametric=function(Y,Z,h,g,nD=6,Quant=TRUE,Fo=0){

	# nodes and weights for the numeric quadrature
	y=gauss.quad(nD)[[1]]*sqrt(2)
	w=gauss.quad(nD)[[2]]/sqrt(pi)

	n=ncol(Y)
	N=nrow(Y)
	
	# specifying starting values
	# estimating parameters for the non-moderated model
	C=coef(ltm(Y~z1))
   	a=C[,2]/1.7
    	b=-a*C[,1]
	
	# person-by-indicator specific parameters
	A=matrix(rep(a,N),ncol=n,byrow=TRUE)
    	B=matrix(rep(b,N),ncol=n,byrow=TRUE)
    
	# slopes and intercepts of the piece-wise regressions connecting the focal points
    	aaA=matrix(ncol=n,nrow=g-1)
    	bbA=matrix(ncol=n,nrow=g-1)
    	aaB=aaA
    	bbB=bbA
    
	# matrix for the parameter estimates at the focal points, the first layer for the intercepts and the second for the slopes

    	Coefs=array(dim=c(2,g,n));
    	for(i in 1:g){
		Coefs[1,i,]=b;
		Coefs[2,i,]=a
	}

	for(gg in 1:2){# repeat the procedure twice
      	for(i in 1:n){# loop over indicators
			if(Quant==TRUE){ # select the focal points for the indicators
            		focal=quantile(Z[,i],seq(1/(2*g),1-1/(2*g),length=g));
            	}
			if(Quant==FALSE){
				focal=Fo[,i]
			}
			bw=h*sd(Z[,i])/(N^(1/5)); # bandwidth for the indicators
            	for(j in 1:g){ # loop over focal points
                		a0=focal[j];

                		# specifying the weights for indicator i and focal point j
				W=exp(-((Z[,i]-a0)/bw)^2/2)
				
				# computing the log-likelihood at the quadrature nodes for indicators k\neq i. It does not depend on the parameters which are optimized at this step, so it is computed outside the optimisation step
                		LL=.C('LLstep1', as.integer(Y),as.double(W),as.double(A),as.double(B),as.double(y),as.double(w),
as.double(numeric(N*nD)),as.integer(N),as.integer(n),as.integer(nD),as.integer(i-1))[[7]]

                		F=function(x){ # -log-likelihood to be maximized
L=.C("LLstep2",as.integer(Y),as.double(W),as.double(x[1]),
as.double(x[2]),as.double(y),as.double(w),as.double(LL),as.double(0),as.integer(N),as.integer(n),as.integer(nD),as.integer(i-1))[[8]]
                  return(-L) }
            		
				# optimising the log-likelihood	
                		Coefs[c(2,1),j,i]=optim(Coefs[c(2,1),j,i],F)[[1]]# updated values of the parameters of indicator i at focal point j
                
             	}
			# compute the slopes and the intercepts of the piece-wise regressions connecting the focal points
            	aaB[,i]=(Coefs[1,-1,i]-Coefs[1,-g,i])/(focal[-1]-focal[-g])
            	bbB[,i]=Coefs[1,-g,i]-aaB[,i]*focal[-g]
            	aaA[,i]=(Coefs[2,-1,i]-Coefs[2,-g,i])/(focal[-1]-focal[-g])
            	bbA[,i]=Coefs[2,-g,i]-aaA[,i]*focal[-g]
            
			# update the values in the matrices of person-by-indicator-specific parameters
           	      A[Z[,i]<focal[1],i]=Coefs[2,1,i]
            	B[Z[,i]<focal[1],i]=Coefs[1,1,i]
            	A[Z[,i]>=focal[g],i]=Coefs[2,g,i]
            	B[Z[,i]>=focal[g],i]=Coefs[1,g,i]
            	for(j in 1:(g-1)){
                		A[(Z[,i]>=focal[j])&(Z[,i]<focal[j+1]),i]=bbA[j,i]+aaA[j,i]*Z[(Z[,i]>=focal[j])&(Z[,i]<focal[j+1]),i]
                		B[(Z[,i]>=focal[j])&(Z[,i]<focal[j+1]),i]=bbB[j,i]+aaB[j,i]*Z[(Z[,i]>=focal[j])&(Z[,i]<focal[j+1]),i]
            	}
	        }
	}

	return(Coefs)		
}

# auxillary functions for sampling the mean and covariance matrix of the normally distributed variables

sample_mu=function(x,s){
    k=ncol(x)
    m0=rep(0,k)
    s0=matrix(c(rep(c(100,rep(0,k)),k-1),100),ncol=k)
    m=mvrnorm(1,solve(solve(s0)+nrow(x)*solve(s))%*%(solve(s0)%*%m0+nrow(x)*solve(s)%*%colMeans(x)),solve(solve(s0)+nrow(x)*solve(s)))
    return(m)
}

sample_S=function (theta,mu)
{
    nS = ncol(theta) ## number of dimensions
    nP = nrow(theta) ## sample size
    mu = matrix(mu, ncol = 1, nrow = nS)
    S = matrix(rep(0, nS^2), ncol = nS, nrow = nS) ### sum of squares matrix
    for (i in 1:nP) {
        x = matrix(theta[i, ], ncol = 1, nrow = nS)
        x = x - mu
        S = S + x %*% t(x)
    }
    s = matrix(c(rep(c(1, rep(0, nS)), (nS - 1)), 1), nS, nS)
    S = S + s
    S = solve(S)
    Z = mvrnorm(n = (nP + nS + 2), mu = c(rep(0, nS)), Sigma = S) ### sampling from the inverse-Wishart (see Gelman et al. 1995)
    T = matrix(rep(0, nS^2), nS, nS)
    for (i in 1:(nP + nS + 2)) {
        z = matrix(Z[i, ], ncol = 1, nrow = nS)
        T = T + z %*% t(z)
    }
    T = solve(T)
    return(T)
}# output: new covariance matrix



# nIt - total number of iterations
# burnin - number of iterations of burnin
# R - degree of the polynomial, if R=0, then a non-moderated model would be estimated
# Cat=FALSE means that a polynomial model is estimated, Cat=TRUE means that a model with a catgorised moderator is estimated
# if Cat==TRUE and Th==0, then median split is used for the moderator, otherwise Th should be a matrix with indicator specific thresholds (columns for indicator, rows for thresholds) 

parametric=function(Y,ZZ=0,nIt,burnin,R,Cat=FALSE,Th=0){
	
	Y=as.matrix(Y)
	n=ncol(Y)	
	N=nrow(Y)

	if(Cat==TRUE){	
		if(length(Th)==1){
			Z=ZZ
			for(i in 1:n){
				Z[,i]=1*(ZZ[,i]>=median(ZZ[,i]))
			}
		}
		if(length(Th)!=1){
			R=nrow(Th)
			Z=array(dim=c(N,n,R))
			for(r in 1:R){
				Z[,,r]=ZZ
				for(i in 1:n){
					th=c(min(ZZ[,i])-1,Th[,i])
					Z[,i,r]=1*(ZZ[,i]>=th[r])&(ZZ[,i]<th[r+1]);
				}
			}
		}
	}	

	if(Cat==FALSE){
		if(R==1){
			Z=apply(ZZ,2,scale)
		}
		if(R>1){
			Z=array(dim=c(N,n,R));
			for(i in 1:n){
				Z[,i,]=apply(poly(ZZ[,i],R),2,scale)
				
			}
		}
	}

	if(R>1){
		Zlist=list()
		for(i in 1:n){
			Zlist[[i]]=Z[,i,]
		}
	}

	# specifying initial values
	M=ltm(Y~z1)
    	a=coef(M)[,2]/1.7
   	b=-coef(M)[,1]*coef(M)[,2]/1.7
    	if(R>0){
		for(r in 1:R){
			a=c(a,rep(0,n))
			b=c(b,rep(0,n))
		}
	}
    	t=factor.scores(M,Y)$score.dat$z1
   	items=matrix(c(a,b),ncol=2*R+2)
    	
	# hyperparameters
	m=rep(0,2*R+2)
    	s=diag(2*R+2)
   
	# continuous augmented data
	X=Y

	# saving the chains of the sampled item parameter values
   	Items=matrix(ncol=n*(2*R+2),nrow=nIt)

				
	if(R==0){Z=0}

	# Starting the Gibbs Sampler
	for(It in 1:nIt){	
		# sample the augmented data and the person parameters (within C)
		tmp<-.C('parametric',as.integer(Y),as.double(X),as.double(c(0,Z)),as.double(t),as.double(a),as.double(b),as.double(numeric(n)),as.double(numeric(n)),as.integer(N),as.integer(n),as.integer(R))
		X=matrix(tmp[[2]],ncol=n)
		t=tmp[[4]]
	
		if(R>1){
			items=t(mapply(Z=Zlist,items=as.data.frame(t(items)),X=as.data.frame(X),FUN=function(Z,items,X){
            		Q=t
				for(j in 1:ncol(Z)){
					Q=cbind(Q,t*Z[,j])
				}
				Q=cbind(Q,rep(1,N))
				for(j in 1:ncol(Z)){
					Q=cbind(Q,Z[,j])
				}
				S=solve(t(Q)%*%Q+solve(s))
            		M=S%*%(t(Q)%*%matrix(X,ncol=1)+solve(s)%*%matrix(m,ncol=1))
      		}))
		}
		if(R==1){
			items=t(mapply(Z=as.data.frame(Z),items=as.data.frame(t(items)),X=as.data.frame(X),FUN=function(Z,items,X){
            		Q=cbind(t,Z*t,rep(1,N),Z)
				S=solve(t(Q)%*%Q+solve(s))
            		M=S%*%(t(Q)%*%matrix(X,ncol=1)+solve(s)%*%matrix(m,ncol=1))
      		}))
		}
		if(R==0){
			items=t(mapply(items=as.data.frame(t(items)),X=as.data.frame(X),FUN=function(items,X){
            		Q=cbind(t,rep(1,N))
				S=solve(t(Q)%*%Q+solve(s))
            		M=S%*%(t(Q)%*%matrix(X,ncol=1)+solve(s)%*%matrix(m,ncol=1))
      		}))
		}

		# making sure that the mean and variance of the person parameter and 0 and 1
		mt=mean(t)
		st=sd(t)
		for(r in 0:R){
			items[,R+1+1+r]=items[,R+1+1+r]+mt*items[,1+r]
			items[,1+r]=items[,1+r]*st
		}
		t=(t-mt)/st
	
		# sampling the mean and the variance of the item parameters
		m=sample_mu(items,s)
      	s=sample_S(items,m)

		a=items[,1:(1+R)]
		b=items[,-c(1:(1+R))]
	
		Items[It,]=c(items)

	}

	# estimates of the parameters
	CoefA=matrix(colMeans(Items[-c(1:burnin),])[1:(n*(1+R))],ncol=n,byrow=TRUE)
   	CoefB=matrix(colMeans(Items[-c(1:burnin),])[-c(1:(n*(1+R)))],ncol=n,byrow=TRUE)

	# person-by-indicator specific parameter matrices A and B
	if(R==0){
		A=matrix(rep(CoefA[1,],N),nrow=N,byrow=TRUE)
		B=matrix(rep(CoefB[1,],N),nrow=N,byrow=TRUE)
	}	
	if(R==1){
		A=mapply(Z=as.data.frame(Z),D=as.data.frame(CoefA),FUN=function(D,Z){D[1]+D[2]*Z})
		B=mapply(Z=as.data.frame(Z),D=as.data.frame(CoefB),FUN=function(D,Z){D[1]+D[2]*Z})
	}
	if(R>1){
		A=mapply(Z=Zlist,D=as.data.frame(CoefA),FUN=function(D,Z){D[1]+Z%*%D[-1]})
		B=mapply(Z=Zlist,D=as.data.frame(CoefB),FUN=function(D,Z){D[1]+Z%*%D[-1]})
	}
	
	# Gauss-Hermite quadrature with 100 nodes is used to approximate the log-likelihood which is needed for model comparisons
	y=gauss.quad(100)[[1]]*sqrt(2)
	w=gauss.quad(100)[[2]]/sqrt(pi)

	# information criteria for model comparison
	LL<-.C('LL',as.integer(Y),as.double(A),as.double(B),as.double(y),as.double(w),as.integer(N),as.integer(n),as.integer(100),as.double(numeric(N)))[[9]]
	nP=n*(2+2*R)
	aic=-2*sum(LL)+2*nP
	bic=-2*sum(LL)+log(N)*nP

	res=list(chains=Items,est=list(lambda=CoefA,nu=CoefB),PbyI=list(lambda=A,nu=B),mAIC=aic,mBIC=bic,nP=nP)

}

